from galaxy.util import rules_dsl


def test_rules():
    for test_case in rules_dsl.get_rules_specification():
        rule_set = rules_dsl.RuleSet(test_case)
        if "initial" in test_case:
            expect_errors = test_case.get("error", False)
            found_error = None
            try:
                initial = test_case["initial"]
                final_data, final_sources = rule_set.apply(initial["data"], initial["sources"])
                expected_final = test_case["final"]
                expected_final_data = expected_final["data"]
                msg = f"Incorrect number of rows, {final_data} != {expected_final_data}"
                assert len(expected_final_data) == len(final_data), msg
                for final_row, expected_final_row in zip(final_data, expected_final_data):
                    msg = f"{final_row} != {expected_final_row}"
                    assert len(final_row) == len(expected_final_row), msg
                    for final_val, expected_final_val in zip(final_row, expected_final_row):
                        assert final_val == expected_final_val, msg
                expected_final_sources = expected_final.get("sources", None)
                if expected_final_sources:
                    msg = f"Incorrect number of sources, {expected_final_sources} != {final_sources}"
                    assert len(expected_final_sources) == len(final_sources), msg
                    for final_source, expected_final_source in zip(final_sources, expected_final_sources):
                        msg = f"{final_source} != {expected_final_source}"
                        assert final_source == expected_final_source, msg
            except AssertionError:
                raise
            except Exception as e:
                found_error = e

            if expect_errors and found_error is None:
                raise AssertionError(f"Did not find expected runtime errors for test case [{test_case}].")
            elif not expect_errors and found_error:
                raise AssertionError(f"Unexpected error when evaluating test case [{test_case}] - [{found_error}].")
        elif "error" in test_case:
            assert rule_set.has_errors, f"rule [{test_case}] does not contain errors"
        else:
            raise Exception(f"Problem with test case definition [{test_case}].")



import ntpath
import os
import posixpath

from galaxy.util import in_directory


def test_in_directory(tmp_path):
    safe_dir = os.path.join(tmp_path, "user")
    os.mkdir(safe_dir)
    good_file = os.path.join(safe_dir, "1")
    with open(good_file, "w") as f:
        f.write("hello")
    assert in_directory(good_file, safe_dir)

    assert not in_directory("/other/file/is/here.txt", safe_dir)

    unsafe_link = os.path.join(safe_dir, "2")
    os.symlink("/other/file/bad.fasta", unsafe_link)
    assert not in_directory(unsafe_link, safe_dir)

    # Test local_path_module parameter
    if os.path == posixpath:
        assert in_directory(good_file, safe_dir, local_path_module=ntpath)
    elif os.path == ntpath:
        assert in_directory(good_file, safe_dir, local_path_module=posixpath)



import os.path
import tempfile

from galaxy.util.checkers import (
    check_html,
    check_image,
)


def test_check_html():
    html_text = '<p>\n<a href="url">Link</a>\n</p>\n'
    assert check_html(html_text, file_path=False)
    # Test a non-HTML binary string
    assert not check_html(b"No HTML here\nSecond line\n", file_path=False)
    with tempfile.NamedTemporaryFile(mode="w") as tmp:
        tmp.write(html_text)
        tmp.flush()
        assert check_html(tmp.name)
    # Test a non-UTF8 binary file
    with tempfile.NamedTemporaryFile(mode="wb") as tmpb:
        tmpb.write(b"\x1f\x8b")
        tmpb.flush()
        assert not check_html(tmpb.name)


def test_check_image():
    for filename, expected in (("1.tiff", True), ("454Score.png", True), ("1.bam", False)):
        path = f"test-data/{filename}"
        assert os.path.exists(path)
        assert check_image(path) is expected



from galaxy.util.search import (
    parse_filters,
    parse_filters_structured,
)


def test_parse_filters():
    result = parse_filters("moo cow", {})
    assert result[0] is None
    assert result[1] == "moo cow"

    result = parse_filters("moo:mooterm cow", {"moo": "mookey"})
    filters = result[0]
    assert filters
    assert filters[0][0] == "mookey"
    assert filters[0][1] == "mooterm"
    assert result[1] == "cow"

    result = parse_filters("moo:'moo term' cow", {"moo": "mookey"})
    filters = result[0]
    assert filters
    assert filters[0][0] == "mookey"
    assert filters[0][1] == "moo term"
    assert result[1] == "cow"

    result = parse_filters("""moo:"moo term" cow""", {"moo": "mookey"})
    filters = result[0]
    assert filters
    assert filters[0][0] == "mookey"
    assert filters[0][1] == "moo term"
    assert result[1] == "cow"

    result = parse_filters("cow moo:'moo term'", {"moo": "mookey"})
    filters = result[0]
    assert filters
    assert filters[0][0] == "mookey"
    assert filters[0][1] == "moo term"
    assert result[1] == "cow"

    result = parse_filters("cow moo:'moo term' other side", {"moo": "mookey"})
    filters = result[0]
    assert filters
    assert filters[0][0] == "mookey"
    assert filters[0][1] == "moo term"
    assert result[1] == "cow other side"


def test_parse_filters_structured():
    result = parse_filters_structured("cow moo:moo other side", {"moo": "mookey"}, preserve_quotes=True)
    filters = result.filter_terms
    assert filters
    assert filters[0].quoted is False
    text_terms = result.text_terms
    assert len(text_terms) == 3
    assert text_terms[0].text == "cow"
    assert text_terms[0].quoted is False
    assert text_terms[1].text == "other"
    assert text_terms[1].quoted is False
    assert text_terms[2].text == "side"
    assert text_terms[2].quoted is False

    result = parse_filters_structured("cow moo:'moo' other side", {"moo": "mookey"}, preserve_quotes=True)
    filters = result.filter_terms
    assert filters
    assert filters[0].quoted is True
    text_terms = result.text_terms
    assert len(text_terms) == 3
    assert text_terms[0].text == "cow"
    assert text_terms[0].quoted is False
    assert text_terms[1].text == "other"
    assert text_terms[1].quoted is False
    assert text_terms[2].text == "side"
    assert text_terms[2].quoted is False

    result = parse_filters_structured("""cow moo:"moo" other side""", {"moo": "mookey"}, preserve_quotes=True)
    filters = result.filter_terms
    assert filters
    assert filters[0].quoted is True

    result = parse_filters_structured("cow moo:'moo term' other side", {"moo": "mookey"}, preserve_quotes=True)
    filters = result.filter_terms
    assert filters
    assert filters[0].filter == "mookey"
    assert filters[0].text == "moo term"
    assert filters[0].quoted is True
    assert " ".join(t.text for t in result.text_terms) == "cow other side"

    result = parse_filters_structured("cow 'other side' foo", {"moo": "mookey"}, preserve_quotes=True)
    text_terms = result.text_terms
    assert len(text_terms) == 3
    assert text_terms[0].text == "cow"
    assert text_terms[0].quoted is False
    assert text_terms[1].text == "other side"
    assert text_terms[1].quoted is True
    assert text_terms[2].text == "foo"
    assert text_terms[2].quoted is False



from decimal import Decimal

import pytest

from galaxy.util.json import (
    safe_dumps,
    swap_inf_nan,
)


@pytest.mark.parametrize(
    "val,expected_val",
    [
        (float("inf"), "__Infinity__"),
        (float("-inf"), "__-Infinity__"),
        (float("NaN"), "__NaN__"),
        (Decimal("1"), "1"),
    ],
)
def test_swap_inf_nan(val, expected_val):
    assert swap_inf_nan(val) == expected_val


def test_safe_dumps():
    assert safe_dumps({"a": Decimal("0.1")}) == """{"a": "0.1"}"""



from typing import (
    Any,
    Dict,
    List,
    Optional,
)

from galaxy.exceptions import (
    RequestParameterInvalidException,
    RequestParameterMissingException,
)
from galaxy.util.config_templates import (
    StrictModel,
    TemplateEnvironmentEntry,
    TemplateSecret,
    TemplateVariable,
    TemplateVariableBoolean,
    TemplateVariableInteger,
    TemplateVariablePathComponent,
    TemplateVariableString,
    validate_secrets_and_variables,
)

TEST_TEMPLATE_ID = "test_id"
TEST_TEMPLATE_VERSION = 0


class TestTemplate(StrictModel):
    id: str
    type: str = "test"
    version: int
    variables: Optional[List[TemplateVariable]]
    secrets: Optional[List[TemplateSecret]]
    environment: Optional[List[TemplateEnvironmentEntry]]


def _template_with_variable(variable: TemplateVariable) -> TestTemplate:
    return TestTemplate(
        id=TEST_TEMPLATE_ID,
        version=TEST_TEMPLATE_VERSION,
        variables=[variable],
        secrets=None,
        environment=None,
    )


def _template_with_secret(name: str) -> TestTemplate:
    secret = TemplateSecret(name=name, help="Help for secret.")
    return TestTemplate(
        id=TEST_TEMPLATE_ID,
        version=TEST_TEMPLATE_VERSION,
        variables=[],
        secrets=[secret],
        environment=None,
    )


class TestInstanceDefinition(StrictModel):
    template_id: str
    template_version: int
    variables: Dict[str, Any]
    secrets: Dict[str, str]


def _test_instance_with_variables(variables: Dict[str, Any]) -> TestInstanceDefinition:
    return TestInstanceDefinition(
        template_id=TEST_TEMPLATE_ID,
        template_version=TEST_TEMPLATE_VERSION,
        variables=variables,
        secrets={},
    )


def _test_instance_with_secrets(secrets: Dict[str, str]) -> TestInstanceDefinition:
    return TestInstanceDefinition(
        template_id=TEST_TEMPLATE_ID,
        template_version=TEST_TEMPLATE_VERSION,
        variables={},
        secrets=secrets,
    )


def test_variable_typing_string():
    template = _template_with_variable(TemplateVariableString(name="test_var", help=None, type="string"))
    instance = _test_instance_with_variables({"test_var": "moocow"})
    validate_secrets_and_variables(instance, template)

    instance = _test_instance_with_variables({"test_var": 5})
    e = assert_validation_throws(instance, template)
    assert isinstance(e, RequestParameterInvalidException)

    instance = _test_instance_with_variables({"test_var": False})
    e = assert_validation_throws(instance, template)
    assert isinstance(e, RequestParameterInvalidException)


def test_variable_typing_boolean():
    template = _template_with_variable(TemplateVariableBoolean(name="test_var", help=None, type="boolean"))
    instance = _test_instance_with_variables({"test_var": False})
    validate_secrets_and_variables(instance, template)

    instance = _test_instance_with_variables({"test_var": 0})
    e = assert_validation_throws(instance, template)
    assert isinstance(e, RequestParameterInvalidException)

    instance = _test_instance_with_variables({"test_var": "false"})
    e = assert_validation_throws(instance, template)
    assert isinstance(e, RequestParameterInvalidException)


def test_variable_typing_int():
    template = _template_with_variable(TemplateVariableInteger(name="test_var", help=None, type="integer"))
    instance = _test_instance_with_variables({"test_var": 6})
    validate_secrets_and_variables(instance, template)

    instance = _test_instance_with_variables({"test_var": False})
    e = assert_validation_throws(instance, template)
    assert isinstance(e, RequestParameterInvalidException)

    instance = _test_instance_with_variables({"test_var": "six"})
    e = assert_validation_throws(instance, template)
    assert isinstance(e, RequestParameterInvalidException)


def test_variable_typing_path_component():
    template = _template_with_variable(TemplateVariablePathComponent(name="test_var", help=None, type="path_component"))
    instance = _test_instance_with_variables({"test_var": "simple_directory"})
    validate_secrets_and_variables(instance, template)

    instance = _test_instance_with_variables({"test_var": 0})
    e = assert_validation_throws(instance, template)
    assert isinstance(e, RequestParameterInvalidException)

    instance = _test_instance_with_variables({"test_var": False})
    e = assert_validation_throws(instance, template)
    assert isinstance(e, RequestParameterInvalidException)

    instance = _test_instance_with_variables({"test_var": "../simple_directory"})
    e = assert_validation_throws(instance, template)
    assert isinstance(e, RequestParameterInvalidException)

    instance = _test_instance_with_variables({"test_var": "/simple_directory"})
    e = assert_validation_throws(instance, template)
    assert isinstance(e, RequestParameterInvalidException)

    instance = _test_instance_with_variables({"test_var": "simple_directory", "extra": "4"})
    e = assert_validation_throws(instance, template)
    assert isinstance(e, RequestParameterInvalidException)


def test_variable_missing():
    template = _template_with_variable(TemplateVariablePathComponent(name="test_var", help=None, type="path_component"))
    instance = _test_instance_with_variables({})
    e = assert_validation_throws(instance, template)
    assert isinstance(e, RequestParameterMissingException)


def test_secret_missing():
    template = _template_with_secret("mysecret")
    instance = _test_instance_with_secrets({})
    e = assert_validation_throws(instance, template)
    assert isinstance(e, RequestParameterMissingException)


def test_extra_secret():
    template = _template_with_secret("mysecret")
    instance = _test_instance_with_secrets({"mysecret": "myvalue"})
    validate_secrets_and_variables(instance, template)

    instance = _test_instance_with_secrets({"mysecret": "myvalue", "extrasecret": "extravalue"})
    e = assert_validation_throws(instance, template)
    assert isinstance(e, RequestParameterInvalidException)


def assert_validation_throws(instance: TestInstanceDefinition, template: TestTemplate) -> Exception:
    try:
        validate_secrets_and_variables(instance, template)
    except Exception as e:
        return e
    raise AssertionError("Expected validation error did not occur.")



import os

from galaxy.util.commands import new_clean_env


def test_new_clean_env() -> None:
    saved_environ = os.environ.copy()
    os.environ["FOO"] = "foo"
    os.environ.pop("TMPDIR", None)
    try:
        clean_env = new_clean_env()
    finally:
        os.environ.clear()
        os.environ.update(saved_environ)
    assert "FOO" not in clean_env
    for k in ("HOME", "PATH"):
        if k in saved_environ:
            assert clean_env[k] == saved_environ[k]
    assert clean_env["LC_CTYPE"].endswith("UTF-8")
    assert clean_env["TMPDIR"]



import os

import pytest

from galaxy.exceptions import InvalidFileFormatError
from galaxy.util import properties
from galaxy.util.properties import (
    get_from_env,
    nice_config_parser,
    read_properties_from_file,
)

KEY1, KEY2, KEY3, KEY4, KEY5, KEY6 = "k1", "k2", "k3", "k4", "k5", "k6"
VAL1, VAL2, VAL3, VAL4, VAL5, VAL6 = 1, 2, 3, 4, 5, 6
OTHER_SECTION = "other"


def test_get_from_env(monkeypatch):
    prefixes = ["pre1_", "pre2_", "pre3_"]

    monkeypatch.setenv("pre1_a", "a-is-set")
    monkeypatch.setenv("pre2_b", "b-is-set")

    assert get_from_env("a", prefixes, "a-default") == "a-is-set"  # selected from first prefix
    assert get_from_env("b", prefixes, "b-default") == "b-is-set"  # selected from second prefix
    assert get_from_env("c", prefixes, "c-default") == "c-default"  # default


def test_get_from_env__no_default(monkeypatch):
    assert get_from_env("a", ["some", "random", "prefixes"]) is None


def test_get_from_env__empty_prefix_ignored(monkeypatch):
    monkeypatch.setenv("a", "a-without-prefix-is-set")
    assert os.getenv("a") == "a-without-prefix-is-set"

    assert get_from_env("a", ["", None], "a-default") == "a-default"  # type: ignore[list-item]


@pytest.fixture
def mock_properties(monkeypatch):
    # Keys in defaults dict and other dicts should not intersect. The tests assume
    # that len(dict1.update(dict2)) == len(dict1) + len(dict2). So, for example, if
    # defaults = {foo: 1, bar:2} and other_section = {foo:3}, the final dict will have len=2.
    def mock_default_properties(path):
        return {KEY1: VAL1, KEY2: VAL2}

    def mock_read_from_yaml(path):
        return {"galaxy": {KEY3: VAL3, KEY4: VAL4}, OTHER_SECTION: {KEY5: VAL5, KEY6: VAL6}}

    class MockConfigParser:
        def __init__(self):
            self._items = {"app:main": [(KEY3, VAL3), (KEY4, VAL4)], OTHER_SECTION: [(KEY5, VAL5), (KEY6, VAL6)]}

        def has_section(self, section):
            return section in self._items.keys()

        def items(self, section):
            return self._items[section] + self.defaults()

        def defaults(self):
            return [(KEY1, VAL1), (KEY2, VAL2)]

    def mock_nice_config_parser(path):
        return MockConfigParser()

    monkeypatch.setattr(properties, "_read_from_yaml_file", mock_read_from_yaml)
    monkeypatch.setattr(properties, "nice_config_parser", mock_nice_config_parser)
    monkeypatch.setattr(properties, "__default_properties", mock_default_properties)


def test_read_galaxy_properties_from_yaml(mock_properties):
    file = "foo.yaml"
    result = read_properties_from_file(file)

    assert len(result) == 4
    assert result[KEY1] == VAL1
    assert result[KEY2] == VAL2
    assert result[KEY3] == VAL3
    assert result[KEY4] == VAL4


def test_read_other_properties_from_yaml(mock_properties):
    file = "foo.yaml"
    result = read_properties_from_file(file, config_section=OTHER_SECTION)

    assert len(result) == 4
    assert result[KEY1] == VAL1
    assert result[KEY2] == VAL2
    assert result[KEY5] == VAL5
    assert result[KEY6] == VAL6


def test_get_default_properties_from_yaml(mock_properties):
    file = "foo.yaml"
    result = read_properties_from_file(file, config_section="section's not here, man")

    assert len(result) == 2
    assert result[KEY1] == VAL1
    assert result[KEY2] == VAL2


def test_read_galaxy_properties_from_ini(mock_properties):
    file = "foo.ini"
    result = read_properties_from_file(file)

    assert len(result) == 4
    assert result[KEY1] == VAL1
    assert result[KEY2] == VAL2
    assert result[KEY3] == VAL3
    assert result[KEY4] == VAL4


def test_read_other_properties_from_ini(mock_properties):
    file = "foo.ini"
    result = read_properties_from_file(file, config_section=OTHER_SECTION)

    assert len(result) == 4
    assert result[KEY1] == VAL1
    assert result[KEY2] == VAL2
    assert result[KEY5] == VAL5
    assert result[KEY6] == VAL6


def test_get_default_properties_from_ini(mock_properties):
    file = "foo.ini"
    result = read_properties_from_file(file, config_section="section's not here, man")

    assert len(result) == 2
    assert result[KEY1] == VAL1
    assert result[KEY2] == VAL2


def test_dont_read_properties_from_other_format(mock_properties):
    file = "invalid"
    with pytest.raises(InvalidFileFormatError):
        read_properties_from_file(file)


def test_nice_config_parser(tmp_path):
    conf_path = tmp_path / "config.ini"
    conf_path.write_text(
        """
[main]
foo_path = %(here)s/cow
FOO_PATH123 = hackedtonotchangecase
config_file = %(__file__)s
"""
    )
    config_parser = nice_config_parser(conf_path)
    foo_path = config_parser.get("main", "foo_path")
    assert foo_path == str(tmp_path / "cow")
    assert config_parser.get("main", "FOO_PATH123") == "hackedtonotchangecase"
    assert config_parser.get("main", "config_file") == str(conf_path)


def test_nice_config_parser_custom_exceptions(tmp_path):
    conf_path = tmp_path / "config.ini"
    conf_path.write_text(
        """
[main]
foo_var = %(foobar)s/cow
"""
    )
    config_parser = nice_config_parser(conf_path)
    exc = None
    try:
        config_parser.get("main", "foo_var")
    except Exception as e:
        exc = e
    assert exc is not None
    assert f"Error in file {conf_path}" in str(exc)
    assert "Bad value substitution: option 'foo_var' in section 'main' contains an interpolation key" in str(exc)



import pytest
import requests
import responses
from werkzeug.wrappers.response import Response

from galaxy.util import url_get


@responses.activate
def test_get_url_ok():
    url = "https://toolshed.g2.bx.psu.edu/"
    responses.add(responses.GET, url, body="OK", status=200)
    text = url_get(url)
    assert text == "OK"


@responses.activate
def test_get_url_forbidden():
    url = "https://toolshed.g2.bx.psu.edu/"
    responses.add(responses.GET, url, body="Forbidden", status=403)
    with pytest.raises(requests.exceptions.HTTPError) as excinfo:
        url_get(url)
    assert "403 Client Error: Forbidden for url: https://toolshed.g2.bx.psu.edu/" in str(excinfo)


def test_get_url_retry_after(httpserver):
    attempts = []

    def retry_handler(request):
        attempts.append(request)
        if len(attempts) < 4:
            return Response("try again later", status=429, content_type="text/plain")
        else:
            return Response("ok", status=200, content_type="text/plain")

    httpserver.expect_request("/429").respond_with_handler(retry_handler)
    url = httpserver.url_for("/429")
    with pytest.raises(requests.exceptions.RetryError):
        url_get(url, max_retries=1, backoff_factor=0.01)
    assert len(attempts) == 2
    response = url_get(url, max_retries=2, backoff_factor=0.01)
    assert len(attempts) == 4
    assert response == "ok"



from galaxy.util.unittest import TestCase


class TestTestCase(TestCase):
    j: str

    @classmethod
    def setUpClass(cls):
        cls.j = "foo"

    def setUp(self):
        self.i = 1

    def test_setUpClass(self):
        assert self.j == "foo"

    def test_setUp(self):
        assert self.i == 1

    def test_assertRaises(self):
        with self.assertRaises(ZeroDivisionError):
            1 / 0  # noqa: B018

    def test_assertRaisesRegex(self):
        with self.assertRaisesRegex(ZeroDivisionError, "^division .* zero"):
            1 / 0  # noqa: B018



import errno
import os
import tempfile
from enum import Enum
from io import StringIO
from pathlib import Path
from typing import Dict

import pytest

from galaxy import util
from galaxy.util.json import safe_loads

SECTION_XML = """<?xml version="1.0" ?>
<section id="fasta_fastq_manipulation" name="Fasta Fastq Manipulation" version="">
    <tool file="toolshed.g2.bx.psu.edu/repos/peterjc/seq_filter_by_id/fb1313d79396/seq_filter_by_id/tools/seq_filter_by_id/seq_filter_by_id.xml" guid="toolshed.g2.bx.psu.edu/repos/peterjc/seq_filter_by_id/seq_filter_by_id/0.2.5">
        <tool_shed>
            toolshed.g2.bx.psu.edu
        </tool_shed>
    </tool>
</section>
"""


def test_strip_control_characters():
    s = "\x00bla"
    assert util.strip_control_characters(s) == "bla"


def test_parse_xml_string():
    section = util.parse_xml_string(SECTION_XML)
    _verify_section(section)


def test_parse_xml_file():
    with tempfile.NamedTemporaryFile(mode="w") as tmp:
        tmp.write(SECTION_XML)
        tmp.flush()
        section = util.parse_xml(tmp.name).getroot()
    _verify_section(section)


def _verify_section(section):
    tool = next(iter(section))
    assert sorted(tool.items()) == [
        (
            "file",
            "toolshed.g2.bx.psu.edu/repos/peterjc/seq_filter_by_id/fb1313d79396/seq_filter_by_id/tools/seq_filter_by_id/seq_filter_by_id.xml",
        ),
        ("guid", "toolshed.g2.bx.psu.edu/repos/peterjc/seq_filter_by_id/seq_filter_by_id/0.2.5"),
    ]
    assert next(iter(tool)).text == "toolshed.g2.bx.psu.edu"


def test_xml_to_string():
    section = util.parse_xml_string(SECTION_XML)
    s = util.xml_to_string(section)
    assert len(s.split("\n")) == 1


def test_xml_to_string_pretty():
    section = util.parse_xml_string(SECTION_XML)
    s = util.xml_to_string(section, pretty=True)
    PRETTY = """<?xml version="1.0" ?>
<section id="fasta_fastq_manipulation" name="Fasta Fastq Manipulation" version="">
    <tool file="toolshed.g2.bx.psu.edu/repos/peterjc/seq_filter_by_id/fb1313d79396/seq_filter_by_id/tools/seq_filter_by_id/seq_filter_by_id.xml" guid="toolshed.g2.bx.psu.edu/repos/peterjc/seq_filter_by_id/seq_filter_by_id/0.2.5">
        <tool_shed>toolshed.g2.bx.psu.edu</tool_shed>
    </tool>
</section>"""
    assert s == PRETTY


def test_parse_xml_enoent():
    with tempfile.NamedTemporaryFile() as temp:
        path = temp.name
    with pytest.raises(IOError) as excinfo:
        util.parse_xml(path)
    assert excinfo.value.errno == errno.ENOENT


def test_clean_multiline_string():
    x = util.clean_multiline_string(
        """
        a
        b
        c
"""
    )
    assert x == "a\nb\nc\n"


def test_iter_start_of_lines():
    assert list(util.iter_start_of_line(StringIO("\n1\n\n12\n123\n1234\n"), 1)) == ["\n", "1", "\n", "1", "1", "1"]


def test_safe_loads():
    d: Dict[str, str] = {}
    rval = safe_loads(d)
    assert rval == d
    assert rval is not d
    rval["foo"] = "bar"
    assert "foo" not in d
    s = '{"foo": "bar"}'
    assert safe_loads(s) == {"foo": "bar"}


def test_in_packages(monkeypatch):
    util_path = Path(util.__file__).parent
    assert util.in_packages() == (not str(util_path).endswith("lib/galaxy/util"))


def test_galaxy_directory(monkeypatch):
    galaxy_dir = util.galaxy_directory()
    assert os.path.isabs(galaxy_dir)
    assert os.path.isfile(os.path.join(galaxy_dir, "run.sh"))


def test_listify() -> None:
    assert util.listify(None) == []
    assert util.listify(False) == []
    assert util.listify(True) == [True]
    assert util.listify("foo") == ["foo"]
    assert util.listify("foo, bar") == ["foo", " bar"]
    assert util.listify("foo, bar", do_strip=True) == ["foo", "bar"]
    list_ = [1, 2, 3]
    assert util.listify(list_) is list_
    assert util.listify((1, 2, 3)) == [1, 2, 3]
    s = {1, 2, 3}
    assert util.listify(s) == [s]
    d = {"a": 1, "b": 2, "c": 3}
    assert util.listify(d) == [d]
    o = object()
    assert util.listify(o) == [o]


def test_enum_values():
    class Stuff(str, Enum):
        A = "a"
        C = "c"
        B = "b"

    assert util.enum_values(Stuff) == ["a", "c", "b"]


DOI_VALID_VALUES = [
    "https://doi.org/10.1234/42",
    "doi.org/10.1234/42",
    "doi:10.1234/42",
    "doi:10.1234567890/42",  # longer prefix
    "doi:10.1234/42ab:%&*$//crazy-suffix/%/&/",
    "doi:10.1234/aa",
]


@pytest.mark.parametrize("input", DOI_VALID_VALUES)
def test_validate_doi_pass(input):
    assert util.validate_doi(input)


DOI_INVALID_VALUES = [
    "http://doi.org/10.1234/42",
    "invalid:10.1234/42",
    "doi:11.1234/42",
    "doi:101234/42",
    "doi:10. 1234/42",
    "doi:10.abc/42",
    "doi:10.1234/ 42",
    "doi:10.1234/42/a b",
]


@pytest.mark.parametrize("input", DOI_INVALID_VALUES)
def test_validate_doi_fail(input):
    assert not util.validate_doi(input)


def test_validate_doi_fail_too_long():
    long_suffix = "a" * (util.DOI_MAX_LENGTH - 12)
    doi = f"doi:10.1000/{long_suffix}"
    assert util.validate_doi(doi)
    assert not util.validate_doi(doi + "a")  # Increase length by 1 past max limit
